import React from 'react';
import ReactDOM from 'react-dom';
import PubSub from 'pubsub-js';
// import request from 'request-promise-native';
import { LatLng } from 'leaflet';
import { osmAuth } from 'osm-auth';
import Body from './Body';
import VectorDataManager from '../ctrl/VectorDataManager';
import PresetsManager from '../ctrl/PresetsManager';
// import ImageryManager from '../ctrl/ImageryManager';
import AccountAuth from '../model/AccountAuth';


/**
 * App is the application starter.
 * It creates mostly everything.
 */
export default class App {
	static origin;
	constructor(params) {
		this.el = params.el
		this.id = params.id
		this.url = params.url

		if (params.origin) {
			App.origin = new LatLng(params.origin[0], params.origin[1])
		} else {
			App.origin = new LatLng(0, 0)
		}

		this.state = {
			center: params.center || { lat: 39.9087, lng: 116.3975 }, // { lat: 39.9087, lng: 116.3975 } undefined
			zoom: params.zoom || 18,
			minZoom: params.minZoom || 8,
			maxZoom: params.maxZoom || 22,
			maxBounds: params.maxBounds || null,
		};
		this.child = React.createRef();
	}

	init() {
		this._initCtrl();
		this._initView();
		this._initIndoorMap();
	}


	/**
	 * Creates controller objects.
	 * @private
	 */
	_initCtrl() {
		window.vectorDataManager = new VectorDataManager();
		window.presetsManager = new PresetsManager();
		// window.imageryManager = new ImageryManager();

		// Preload some data
		window.presetsManager.loadPresets();
		// window.imageryManager.getAvailableImagery();
	}

	/**
	 * Create view components
	 * @private
	 */
	_initView() {
		// ref={child => this.View = child}
		ReactDOM.render(<Body {...this.state} ref={child => this.View = child.map} />, document.getElementById(this.el));
		// document.title = window.EDITOR_NAME;

		// Disabling context menu
		document.oncontextmenu = () => {
			return false;
		};
	}

	/**
	 * 加载数据库geojson数据
	 * @private
	 */
	_initIndoorMap() {
		new AccountAuth(this.url, this.id);
	}


	/**
	 * Launches authentication process
	 * @private
	 */
	_initAuth() {
		const opts = {
			apiUrl: window.CONFIG.osm_api_url,
			url: window.CONFIG.osm_api_url,
			client_id: window.CONFIG.oauth_consumer_key,
			redirect_uri: window.EDITOR_URL,
			scope: "read_prefs write_api",
			singlepage: true
		};

		window.editor_user_auth = osmAuth(opts);

		const params = this._readURLParams(window.location.href);

		if (params.code) {
			window.editor_user_auth.authenticate(() => {
				localStorage.setItem("oauth_token", params.code);
				window.history.replaceState({}, "", window.location.href.replace(/\?.*/, ""));
			});
		}
		else if (localStorage.getItem("oauth_token")) {
			window.editor_user_auth.bootstrapToken(localStorage.getItem("oauth_token"), () => {
				this._checkAuth();
			});
		}
		else {
			//Check if we receive auth token
			this._checkAuth();
			this.authWait = setInterval(this._checkAuth.bind(this), 100);
		}
	}


	/**
	 * Parse URL parameters
	 * @private
	 */
	_readURLParams(str) {
		const u = str.split('?');

		if (u.length > 1) {
			const p = u[1].split('#')[0];

			return p.split('&').filter(function (pair) {
				return pair !== '';
			}).reduce(function (obj, pair) {
				var parts = pair.split('=');
				obj[decodeURIComponent(parts[0])] = (null === parts[1]) ?
					'' : decodeURIComponent(parts[1]);
				return obj;
			}, {});
		}
		else {
			return {};
		}
	}


	// /** * 更新标记位置 * * 
	//  * @param {Object} data - 根据data.type类型，将不同的标记按照data.data移动 * 
	// */
	// moveMarker(data) {
	// 	if (data.type === "card") {
	// 		let cards = data.data.map(card => {
	// 			let latlng = MapUtil.distancePointToLatLng(App.origin, card.x, card.y);
	// 			return new LocationCard(latlng.lat, latlng.lng, card.loc_time, card.milli_time, card.card_id, card.info, card.src);
	// 		});
	// 		PubSub.publish('body.data.update', cards);
	// 	}
	// }

	/** * 加载地图文件 * * 
	 * @param {String} url - 地图文件路径 * 
	*/
	loadMap(url) {
		fetch(url)
			.then(res => res.text())
			.then(data => {
				// 处理获取的数据
				let json = JSON.parse(data);
				PubSub.publish("body.draw.loaded", json);
			})
			.catch(error => {
				// 处理错误
				console.error('There was a problem with the fetch operation:', error);
			});
	}

	/** 添加地图事件
	 * 
	 */
	on(eventName, callback) {
		PubSub.publish("map.event.bind", {
			eventName: eventName, callback: callback
		});
	}
	off(eventName, callback) {
		PubSub.publish("map.event.unbind", {
			eventName: eventName, callback: callback
		});
	}
}